#include "stm32f4_discovery.h"
#include "stm32f4xx_conf.h"
#include "stm32f4_discovery_lis302dl.h"
#include "usbh_core.h"
#include "usbh_hid_core.h"
#include "usbh_usr.h"
#include "usb_hcd_int.h"

#include "usart.h"
#include "adb.h"
#include "mac.h"


extern void USB_OTG_BSP_TimerIRQ (void);

void USART2_Callback(uint8_t c);

__ALIGN_BEGIN USBH_HOST                     USB_Host __ALIGN_END ;
__ALIGN_BEGIN USB_OTG_CORE_HANDLE           USB_OTG_Core_dev __ALIGN_END ;
__IO struct adb adb;
__IO unsigned long tenustimer = 0;
__IO unsigned long modifierkeys = 0xFF;
__IO unsigned char usbinitted = 0;
extern __IO uint8_t adbmouseid;

void SysTick_Handler(void) {
	tenustimer++;
	MacHandler();
	ADBTick_Handler();
}

void TIM2_IRQHandler(void) {
	USB_OTG_BSP_TimerIRQ();
}

void OTG_FS_IRQHandler(void) {
	USBH_OTG_ISR_Handler(&USB_OTG_Core_dev);
}

void EXTI1_IRQHandler(void)
{
	if(EXTI_GetITStatus(EXTI_Line1) != RESET) {
		USB_Host.usr_cb->OverCurrentDetected();
		EXTI_ClearITPendingBit(EXTI_Line1);
	}
}

uint32_t LIS302DL_TIMEOUT_UserCallback(void)
{
	USART2_SendString("\n\rtimeout\n\r");
	return 0;
}

int main(void)
{
	unsigned long timeout;
	LIS302DL_InitTypeDef  LIS302DL_InitStruct;
	LIS302DL_FilterConfigTypeDef LIS302DL_FilterStruct;
	int16_t buffer[3];
	int16_t X_Offset, Y_Offset, Z_Offset;
	int16_t oldx = 0, oldy = 0;

	ADBInit();
	MacInit();
	STM_EVAL_LEDInit(LED6);
	STM_EVAL_LEDOff(LED6);

	STM_EVAL_PBInit(BUTTON_USER, BUTTON_MODE_GPIO);


	LIS302DL_InitStruct.Power_Mode = LIS302DL_LOWPOWERMODE_ACTIVE;
	LIS302DL_InitStruct.Output_DataRate = LIS302DL_DATARATE_100;
	LIS302DL_InitStruct.Axes_Enable = LIS302DL_XYZ_ENABLE;
	LIS302DL_InitStruct.Full_Scale = LIS302DL_FULLSCALE_2_3;
	//LIS302DL_InitStruct.Full_Scale = LIS302DL_FULLSCALE_9_2;
	LIS302DL_InitStruct.Self_Test = LIS302DL_SELFTEST_NORMAL;
	LIS302DL_Init(&LIS302DL_InitStruct);
	SysTick_Config(SystemCoreClock/100000);

	timeout = tenustimer + 30;
	while(tenustimer < timeout);

	//LIS302DL_FilterStruct.HighPassFilter_Data_Selection = LIS302DL_FILTEREDDATASELECTION_OUTPUTREGISTER;
	LIS302DL_FilterStruct.HighPassFilter_Data_Selection = LIS302DL_FILTEREDDATASELECTION_BYPASSED;
	LIS302DL_FilterStruct.HighPassFilter_CutOff_Frequency = LIS302DL_HIGHPASSFILTER_LEVEL_1;
	LIS302DL_FilterStruct.HighPassFilter_Interrupt = LIS302DL_HIGHPASSFILTERINTERRUPT_1_2;
	//LIS302DL_FilterStruct.HighPassFilter_Interrupt = LIS302DL_HIGHPASSFILTERINTERRUPT_OFF;
	LIS302DL_FilterConfig(&LIS302DL_FilterStruct);

	adb.device[adbmouseid].devid = adbmouseid << 4;
	adb.device[adbmouseid].reg3 = 0x6301;
	adb.activedevs |= 1<<adbmouseid;

	USART2_Init(USART2_Callback);
	USART2_SendString("USB adapter starting up\n\r");
	USBH_Init(&USB_OTG_Core_dev, USB_OTG_FS_CORE_ID, &USB_Host, &HID_cb, &USR_Callbacks);

	while(!usbinitted) {
		uint8_t update = 0;
		USBH_Process(&USB_OTG_Core_dev, &USB_Host);


		if(adb.senddata & (1<<adbmouseid)) continue;

		LIS302DL_Read((uint8_t*)buffer, LIS302DL_OUT_X_ADDR, 6);
		X_Offset = buffer[0];
		Y_Offset = buffer[1];
		Z_Offset = buffer[2];

#define APPROXTHRESH 0
#define APPROXEQ(old, new) ((((new) > (old)) && (((new) - (old)) <= APPROXTHRESH)) || (((old) >= (new)) && (((old) - (new)) <= APPROXTHRESH)))

        	adb.device[adbmouseid].reg0 = 0;
        	adb.device[adbmouseid].reg0 |= 0x80; // always 1
		if(!STM_EVAL_PBGetState(BUTTON_USER))
			adb.device[adbmouseid].reg0 |= 0x8000;
		if(!APPROXEQ(oldx, X_Offset)) {
			//USART2_SendString(" X: ");
			//USART2_SendHexByte(X_Offset);
			int8_t xupd = ~X_Offset;
			xupd--;
        		adb.device[adbmouseid].reg0 |= (xupd & 0x80) >> 1;
        		adb.device[adbmouseid].reg0 |= (xupd) & 0x3F;
			update = 1;
		}
		oldx = X_Offset;

		if(!APPROXEQ(oldy, Y_Offset)) {
			//USART2_SendString(" Y: ");
			//USART2_SendHexByte(Y_Offset);
			int8_t yupd =~Y_Offset;
			adb.device[adbmouseid].reg0 |= (yupd & 0x80) << 7;
			adb.device[adbmouseid].reg0 |= ((yupd) & 0x3F) << 8;
			update = 1;
		}
		oldy = Y_Offset;

		if(update) {
        		adb.senddata |= (1<<adbmouseid);
		}
#if 0
		if(Z_Offset != oldz) {
			USART2_SendString(" Z: ");
			USART2_SendHexByte(Z_Offset);
			oldz = Z_Offset;
		}
#endif

		//USART2_SendHexByte(ydata>>8);
		//USART2_SendHexByte(ydata);
	}
	USART2_SendString("USB Device attached, beginning normal operation\n\r");
	while(1) {
		USBH_Process(&USB_OTG_Core_dev, &USB_Host);
	}
}

void USART2_Callback(uint8_t c)
{
	USART2_SendHexByte(c);
}

