/**
  ******************************************************************************
  * @file    usbh_usr.c
  * @author  MCD Application Team
  * @version V2.1.0
  * @date    19-March-2012
  * @brief   This file includes the user application layer
  ******************************************************************************
  * @attention
  *
  * <h2><center>&copy; COPYRIGHT 2012 STMicroelectronics</center></h2>
  *
  * Licensed under MCD-ST Liberty SW License Agreement V2, (the "License");
  * You may not use this file except in compliance with the License.
  * You may obtain a copy of the License at:
  *
  *        http://www.st.com/software_license_agreement_liberty_v2
  *
  * Unless required by applicable law or agreed to in writing, software 
  * distributed under the License is distributed on an "AS IS" BASIS, 
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  * See the License for the specific language governing permissions and
  * limitations under the License.
  *
  ******************************************************************************
  */ 

/* Includes ------------------------------------------------------------------*/
#include "usbh_usr.h"
#include "usbh_hid_mouse.h"
#include "usbh_hid_keybd.h"
#include "stm32f4_discovery.h"

#include "usart.h"
#include "adb.h"
#include "mac.h"

#ifndef USE_USB_OTG_FS
#error "USE_USB_OTG_FS expected to be defined"
#endif

#define kBuffSize  32
extern __IO struct adb adb;
extern __IO uint8_t adbmouseid;
extern __IO uint8_t adbkbid;
extern __IO uint8_t adbkey[kBuffSize];
extern __IO uint8_t adbstart;
extern __IO uint8_t adbend;
extern __IO unsigned char usbinitted;
extern uint8_t modifierkeys;
uint8_t USBtoADBScancode(uint8_t code, uint8_t keyup);
void ADBSendKey(uint8_t maccode);
void ADBHandleMeta(uint8_t metakey, uint8_t keyup);

uint32_t USBtoMacScancode(uint8_t code, uint8_t keyup);
void MacHandleMeta(uint8_t metakey, uint8_t keyup);
void MacSendKey(uint32_t maccode);

extern __IO uint16_t mousex;
extern __IO uint16_t mousey;

//extern USB_OTG_CORE_HANDLE           USB_OTG_Core_dev;
/*  Points to the DEVICE_PROP structure of current device */
/*  The purpose of this register is to speed up the execution */

USBH_Usr_cb_TypeDef USR_Callbacks =
{
  USBH_USR_Init,
  USBH_USR_DeInit,
  USBH_USR_DeviceAttached,
  USBH_USR_ResetDevice,
  USBH_USR_DeviceDisconnected,
  USBH_USR_OverCurrentDetected,
  USBH_USR_DeviceSpeedDetected,
  USBH_USR_Device_DescAvailable,
  USBH_USR_DeviceAddressAssigned,
  USBH_USR_Configuration_DescAvailable,
  USBH_USR_Manufacturer_String,
  USBH_USR_Product_String,
  USBH_USR_SerialNum_String,
  USBH_USR_EnumerationDone,
  USBH_USR_UserInput,
  NULL,
  USBH_USR_DeviceNotSupported,
  USBH_USR_UnrecoveredError
};

const char MSG_HOST_INIT[]          = "> Host Library Initialized\n\r";
const char MSG_DEV_ATTACHED[]       = "> Device Attached\n\r";
const char MSG_DEV_DISCONNECTED[]   = "> Device Disconnected\n\r";
const char MSG_DEV_ENUMERATED[]     = "> Enumeration completed\n\r";
const char MSG_DEV_HIGHSPEED[]      = "> High speed device detected\n\r";
const char MSG_DEV_FULLSPEED[]      = "> Full speed device detected\n\r";
const char MSG_DEV_LOWSPEED[]       = "> Low speed device detected\n\r";
const char MSG_DEV_ERROR[]          = "> Device fault \n\r";

const char MSG_MSC_CLASS[]          = "> Mass storage device connected\n\r";
const char MSG_HID_CLASS[]          = "> HID device connected\n\r";

const char USB_HID_MouseStatus[]    = "> HID Demo Device : Mouse\n\r";
const char USB_HID_KeybrdStatus[]   = "> HID Demo Device : Keyboard\n\r";
const char MSG_UNREC_ERROR[]        = "> UNRECOVERED ERROR STATE\n\r";

/**
* @brief  USBH_USR_Init 
*         Displays the message on LCD for host lib initialization
* @param  None
* @retval None
*/
void USBH_USR_Init(void)
{
  static uint8_t startup = 0;  

  //USART2_SendString("USBH_USR_Init\n\r");
  
  if(startup == 0 )
  {
    startup = 1;
    //USART2_SendString(" USB OTG FS HID Host\n\r");
  }
}

/**
* @brief  USBH_USR_DeviceAttached 
*         Displays the message on LCD on device attached
* @param  None
* @retval None
*/
void USBH_USR_DeviceAttached(void)
{  
    //USART2_SendString(MSG_DEV_ATTACHED);
}

/**
* @brief  USBH_USR_UnrecoveredError
* @param  None
* @retval None
*/
void USBH_USR_UnrecoveredError (void)
{
    USART2_SendString(MSG_UNREC_ERROR);
}

/**
* @brief  USBH_DisconnectEvent
*         Device disconnect event
* @param  None
* @retval None
*/
void USBH_USR_DeviceDisconnected (void)
{
    USART2_SendString(MSG_DEV_DISCONNECTED);
}

/**
* @brief  USBH_USR_ResetUSBDevice 
*         Reset USB Device
* @param  None
* @retval None
*/
void USBH_USR_ResetDevice(void)
{
  /* Users can do their application actions here for the USB-Reset */
//USART2_SendString("USB Reset device\n\r");
}


/**
* @brief  USBH_USR_DeviceSpeedDetected 
*         Displays the message on LCD for device speed
* @param  Devicespeed : Device Speed
* @retval None
*/
void USBH_USR_DeviceSpeedDetected(uint8_t DeviceSpeed)
{
  if(DeviceSpeed == HPRT0_PRTSPD_HIGH_SPEED)
  {
   //USART2_SendString(MSG_DEV_HIGHSPEED);
  }
  else if(DeviceSpeed == HPRT0_PRTSPD_FULL_SPEED)
  {
//USART2_SendString(MSG_DEV_FULLSPEED);
  }
  else if(DeviceSpeed == HPRT0_PRTSPD_LOW_SPEED)
  {
    //USART2_SendString(MSG_DEV_LOWSPEED);
  }
  else
  {
    USART2_SendString(MSG_DEV_ERROR);
 }
}

/**
* @brief  USBH_USR_Device_DescAvailable 
*         Displays the message on LCD for device descriptor
* @param  DeviceDesc : device descriptor
* @retval None
*/
void USBH_USR_Device_DescAvailable(void *DeviceDesc)
{
//  uint8_t temp[50];
//  USBH_DevDesc_TypeDef *hs;
//  hs = DeviceDesc;
//
//
//  sprintf((char *)temp , "VID : %04Xh\n" , (uint32_t)(*hs).idVendor); /
//  //LCD_UsrLog((void *)temp);
//
//
//  sprintf((char *)temp , "PID : %04Xh\n" , (uint32_t)(*hs).idProduct);
//  //LCD_UsrLog((void *)temp);
	//USART2_SendString("Device descriptor available\n\r");
}

/**
* @brief  USBH_USR_DeviceAddressAssigned 
*         USB device is successfully assigned the Address 
* @param  None
* @retval None
*/
void USBH_USR_DeviceAddressAssigned(void)
{
	//USART2_SendString("Device address assigned\n\r");
  
}


/**
* @brief  USBH_USR_Conf_Desc 
*         Displays the message on LCD for configuration descriptor
* @param  ConfDesc : Configuration descriptor
* @retval None
*/
void USBH_USR_Configuration_DescAvailable(USBH_CfgDesc_TypeDef * cfgDesc,
                                          USBH_InterfaceDesc_TypeDef *itfDesc,
                                          USBH_EpDesc_TypeDef *epDesc)
{
  USBH_InterfaceDesc_TypeDef *id;

  id = itfDesc;

  if((*id).bInterfaceClass  == 0x08)
  {
    //USART2_SendString(MSG_MSC_CLASS);
  }
  else if((*id).bInterfaceClass  == 0x03)
  {
    //USART2_SendString(MSG_HID_CLASS);
  }
}

/**
* @brief  USBH_USR_Manufacturer_String 
*         Displays the message on LCD for Manufacturer String 
* @param  ManufacturerString : Manufacturer String of Device
* @retval None
*/
void USBH_USR_Manufacturer_String(void *ManufacturerString)
{
    USART2_SendString("Manufacturer: ");
    USART2_SendString(ManufacturerString);
    USART2_SendString("\n\r");
  
}

/**
* @brief  USBH_USR_Product_String 
*         Displays the message on LCD for Product String
* @param  ProductString : Product String of Device
* @retval None
*/
void USBH_USR_Product_String(void *ProductString)
{
    USART2_SendString("Product: ");
    USART2_SendString(ProductString);
    USART2_SendString("\n\r");
  
}

/**
* @brief  USBH_USR_SerialNum_String 
*         Displays the message on LCD for SerialNum_String 
* @param  SerialNumString : SerialNum_String of device
* @retval None
*/
void USBH_USR_SerialNum_String(void *SerialNumString)
{
    USART2_SendString("Serial: ");
    USART2_SendString(SerialNumString);
    USART2_SendString("\n\r");
} 

/**
* @brief  EnumerationDone 
*         User response request is displayed to ask for
*         application jump to class
* @param  None
* @retval None
*/
void USBH_USR_EnumerationDone(void)
{
  /* Enumeration complete */
  USART2_SendString(MSG_DEV_ENUMERATED);
} 

/**
* @brief  USBH_USR_DeviceNotSupported
*         Device is not supported
* @param  None
* @retval None
*/
void USBH_USR_DeviceNotSupported(void)
{
  //USART2_SendString("> Device not supported.\n\r");              
  
}  


/**
* @brief  USBH_USR_UserInput
*         User Action for application state entry
* @param  None
* @retval USBH_USR_Status : User response for key button
*/
USBH_USR_Status USBH_USR_UserInput(void)
{
  return USBH_USR_RESP_OK;
} 

/**
* @brief  USBH_USR_OverCurrentDetected
*         Device Overcurrent detection event
* @param  None
* @retval None
*/
void USBH_USR_OverCurrentDetected (void)
{
   USART2_SendString ("Overcurrent detected.\n");
  
}

/**
* @brief  USR_MOUSE_Init
*         Init Mouse window
* @param  None
* @retval None
*/
void USR_MOUSE_Init	(void)
{
	USART2_SendString(USB_HID_MouseStatus); 
	adb.device[adbmouseid].devid = adbmouseid << 4;
	adb.device[adbmouseid].reg3 = 0x6301;
	adb.activedevs |= 1<<adbmouseid;
	STM_EVAL_LEDOn(LED6);
	usbinitted = 1;
}

/**
* @brief  USR_MOUSE_ProcessData
*         Process Mouse data
* @param  data : Mouse data to be displayed
* @retval None
*/
void USR_MOUSE_ProcessData(HID_MOUSE_Data_TypeDef *data)
{
	if(data->y && ((mousey & 0x00FF) == 0)) {
		mousey = data->y & 0x7F;
		if(!(data->y & 0x80)) {
			mousey |= 0x0100;
		} else {
			mousey = ~data->y & 0x7F;
		}
	}
	if(data->x && ((mousex & 0x00FF) == 0)) {
		mousex = data->x & 0x7F;
		if(data->x & 0x80) {
			mousex = ~data->x & 0x7F;
			mousex++;
			mousex |= 0x0100;
		}
	}
	if(data->button & 1) {
		GPIO_ResetBits(GPIOD, kMSESW);
	} else {
		GPIO_SetBits(GPIOD, kMSESW);
	}


	if(adb.senddata & (1<<adbmouseid)) return;
	//USART2_SendString("pd: s\n\r");

	adb.device[adbmouseid].reg0 = 0;
	if(!(data->button & 1)) {
		//USART2_SendString("Mouse click!\n\r");
		adb.device[adbmouseid].reg0 |= 0x8000;
	}
	adb.device[adbmouseid].reg0 |= (data->y & 0x80) << 7;
	adb.device[adbmouseid].reg0 |= (data->y & 0x3F) << 8;
	adb.device[adbmouseid].reg0 |= 0x80; // always 1
	adb.device[adbmouseid].reg0 |= (data->x & 0x80) >> 1;
	adb.device[adbmouseid].reg0 |= data->x & 0x3F;
	//USART2_SendString("Sending mouse data\n\r");
	adb.senddata |= (1<<adbmouseid);
	return;
}

/**
* @brief  USR_KEYBRD_Init
*         Init Keyboard window
* @param  None
* @retval None
*/
void  USR_KEYBRD_Init (void)
{

	USART2_SendString(USB_HID_KeybrdStatus); 
	adb.device[adbkbid].devid = adbkbid << 4;
	adb.device[adbkbid].reg2 = 0xFF00;
	adb.device[adbkbid].reg3 = 0x6202;
	adb.activedevs = 1<<adbkbid;
	STM_EVAL_LEDOn(LED6);
	usbinitted = 1;
	adb.senddata = 0;
}


/**
* @brief  USR_KEYBRD_ProcessData
*         Process Keyboard data
* @param  data : Keyboard data to be displayed
* @retval None
*/
void  USR_KEYBRD_ProcessData (uint8_t *data)
{
	int i;
	static uint8_t keysdown[6];
	static uint8_t metakeys = 0;
	uint8_t metachanges = data[0] ^ metakeys;

	USART2_SendString("Keyboard data: ");
	for(i = 2; i < 8; i++) {
		if(data[i] && (data[i] < 4)) return;
		//USART2_SendHexByte(data[i]);
	}

	for(i = 0; i < 8; i++) {
		if(metachanges & (1<<i)) {
			ADBHandleMeta(1<<i, (metakeys & (1<<i)));
			MacHandleMeta(1<<i, (metakeys & (1<<i)));
		}
	}
	metakeys = data[0];

	for(i = 2; i < 8; i++) {
		if(data[i] != keysdown[i-2]) {
			uint8_t maccode = 0;
			if(data[i]) {
				maccode = USBtoADBScancode(data[i], 0);
			} else {
				maccode = USBtoADBScancode(keysdown[i-2], 1);
			}
			USART2_SendHexByte(maccode);
			ADBSendKey(maccode);

			uint32_t maccode32 = 0;
			if(data[i]) {
				maccode32 = USBtoMacScancode(data[i], 0);
			} else {
				maccode32 = USBtoMacScancode(keysdown[i-2], 1);
			}
			USART2_SendHexByte(maccode32);
			MacSendKey(maccode32);
		}
		keysdown[i-2] = data[i];
	}
	
	USART2_SendString("\n\r");
}

/**
* @brief  USBH_USR_DeInit
*         Deint User state and associated variables
* @param  None
* @retval None
*/
void USBH_USR_DeInit(void)
{
	USART2_SendString("USBH_USR_DeInit\n\r");
	STM_EVAL_LEDOff(LED6);
}

/**
* @}
*/ 

/**
* @}
*/ 

/**
* @}
*/

/**
* @}
*/
/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/


#define MODKEYSET(y) keyup ? (modifierkeys |= (y)) : (modifierkeys &= ~(y))
uint8_t USBtoADBScancode(uint8_t code, uint8_t keyup) {
	uint8_t maccode = 0;

	switch(code) {
		case 0x04: maccode = 0x00; break; // A
		case 0x05: maccode = 0x0B; break; // B
		case 0x06: maccode = 0x08; break; // C
		case 0x07: maccode = 0x02; break; // D
		case 0x08: maccode = 0x0E; break; // E
		case 0x09: maccode = 0x03; break; // F
		case 0x0A: maccode = 0x05; break; // G
		case 0x0B: maccode = 0x04; break; // H
		case 0x0C: maccode = 0x22; break; // I
		case 0x0D: maccode = 0x26; break; // J
		case 0x0E: maccode = 0x28; break; // K
		case 0x0F: maccode = 0x25; break; // L
		case 0x10: maccode = 0x2E; break; // M
		case 0x11: maccode = 0x2D; break; // N
		case 0x12: maccode = 0x1F; break; // O
		case 0x13: maccode = 0x23; break; // P
		case 0x14: maccode = 0x0C; break; // Q
		case 0x15: maccode = 0x0F; break; // R
		case 0x16: maccode = 0x01; break; // S
		case 0x17: maccode = 0x11; break; // T
		case 0x18: maccode = 0x20; break; // U
		case 0x19: maccode = 0x09; break; // V
		case 0x1A: maccode = 0x0D; break; // W
		case 0x1B: maccode = 0x07; break; // X
		case 0x1C: maccode = 0x10; break; // Y
		case 0x1D: maccode = 0x06; break; // Z
		case 0x1E: maccode = 0x12; break; // 1
		case 0x1F: maccode = 0x13; break; // 2
		case 0x20: maccode = 0x14; break; // 3
		case 0x21: maccode = 0x15; break; // 4
		case 0x22: maccode = 0x17; break; // 5
		case 0x23: maccode = 0x16; break; // 6
		case 0x24: maccode = 0x1A; break; // 7
		case 0x25: maccode = 0x1C; break; // 8
		case 0x26: maccode = 0x19; break; // 9
		case 0x27: maccode = 0x1D ;break; // 0
		case 0x28: maccode = 0x24; break; // enter
		case 0x29: maccode = 0x35; break; // ESC
		case 0x2A: maccode = 0x33; break; // bkspc
		case 0x2B: maccode = 0x30; break; // tab
		case 0x2C: maccode = 0x31; break; // space
		case 0x2D: maccode = 0x1B; break; // -
		case 0x2E: maccode = 0x18; break; // =
		case 0x2F: maccode = 0x21; break; // [
		case 0x30: maccode = 0x1E; break; // ]
		case 0x31: maccode = 0x2A; break; // backslash

		case 0x33: maccode = 0x29; break; // ;
		case 0x34: maccode = 0x27; break; // '
		case 0x35: maccode = 0x32; break; // `
		case 0x36: maccode = 0x2B; break; // ,
		case 0x37: maccode = 0x2F; break; // .
		case 0x38: maccode = 0x2C; break; // /
		case 0x39: maccode = 0x39; break; // caps
		case 0x3A: maccode = 0x7A; break; // F1
		case 0x3B: maccode = 0x78; break; // F2
		case 0x3C: maccode = 0x63; break; // F3
		case 0x3D: maccode = 0x76; break; // F4
		case 0x3E: maccode = 0x60; break; // F5
		case 0x3F: maccode = 0x61; break; // F6
		case 0x40: maccode = 0x62; break; // F7
		case 0x41: maccode = 0x64; break; // F8
		case 0x42: maccode = 0x65; break; // F9
		case 0x43: maccode = 0x6D; break; // F10
		case 0x44: maccode = 0x67; break; // F11
		case 0x45: maccode = 0x6F; break; // F12

		case 0x49: maccode = 0x72; break; // Insert/Help
		case 0x4A: maccode = 0x73; break; // Home
		case 0x4B: maccode = 0x74; break; // PG UP
		case 0x4C: maccode = 0x75; break; // Del
		case 0x4D: maccode = 0x77; break; // End
		case 0x4E: maccode = 0x79; break; // PG DN
		case 0x4F: maccode = 0x3C; break;  // right arrow
		case 0x50: maccode = 0x3B; break;  // left arrow
		case 0x51: maccode = 0x3D; break;  // down arrow
		case 0x52: maccode = 0x3E; break;  // up -> keypad up
		
		case 0x54: maccode = 0x4B; break;  // keypad /
		case 0x55: maccode = 0x43; break; // keypad *
		case 0x56: maccode = 0x4E; break; // keypad -
		case 0x57: maccode = 0x45; break; // keypad +
		case 0x58: maccode = 0x4C; break; // keypad enter
		case 0x59: maccode = 0x53; break; // keypad 1
		case 0x5A: maccode = 0x54; break; // keypad 2
		case 0x5B: maccode = 0x55; break; // keypad 3
		case 0x5C: maccode = 0x56; break; // keypad 4
		case 0x5D: maccode = 0x57; break; // keypad 5
		case 0x5E: maccode = 0x58; break; // keypad 6
		case 0x5F: maccode = 0x59; break; // keypad 7
		case 0x60: maccode = 0x5B; break; // keypad 8
		case 0x61: maccode = 0x5C; break; // keypad 9
		case 0x62: maccode = 0x52; break; // keypad 0
		case 0x63: maccode = 0x41; break; // keypad .
	}

	if((!(modifierkeys & kModShift)) && (maccode == 0x35)) {
		MODKEYSET(kModReset);
		if(keyup) {
			GPIO_SetBits(GPIOD, kADBPowerPin);
			USART2_SendString("Power key up\n\r");
			__disable_irq();
			adbkey[adbend] = 0xFF;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0xFF;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0xFF;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0x38;
			adbend = (adbend+1) % kADBBuffSize;
			adb.senddata |= (1<<adbkbid);
			__enable_irq();
		} else {
			GPIO_ResetBits(GPIOD, kADBPowerPin);
			USART2_SendString("Power key down\n\r");
			__disable_irq();
			adbkey[adbend] = 0x38 | 0x80;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0x7F;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0x7F;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0x7F;
			adbend = (adbend+1) % kADBBuffSize;
			adb.senddata |= (1<<adbkbid);
			__enable_irq();
		}
		return 0x7F;
	}

	if(keyup) maccode |= 0x80;

	return maccode;
}

void ADBSendKey(uint8_t maccode) {
	__disable_irq();
	adbkey[adbend] = maccode;
	adbend = (adbend+1) % kBuffSize;
	adb.senddata |= (1<<adbkbid);
	__enable_irq();
}

void ADBHandleMeta(uint8_t metakey, uint8_t keyup) {
	uint8_t maccode = 0;

	switch(metakey) {
	case KBD_LEFT_SHIFT: maccode = 0x38; MODKEYSET(kModShift); break;
	case KBD_RIGHT_SHIFT: maccode = 0x7B; MODKEYSET(kModShift); break;
	case KBD_LEFT_CTRL: maccode = 0x36; MODKEYSET(kModControl); break;
	case KBD_RIGHT_CTRL: maccode = 0x7D; MODKEYSET(kModControl); break;
	case KBD_LEFT_GUI: maccode = 0x37; MODKEYSET(kModCmd); break;
	case KBD_RIGHT_GUI: maccode = 0x37; MODKEYSET(kModCmd); break;
	case KBD_LEFT_ALT: maccode = 0x3A; MODKEYSET(kModOpt); break;
	case KBD_RIGHT_ALT: maccode = 0x7C; MODKEYSET(kModOpt); break;
	default: return;
	}
	if(keyup) maccode |= 0x80;

	ADBSendKey(maccode);
}

uint32_t USBtoMacScancode(uint8_t code, uint8_t keyup) {
	uint8_t maccode = 0;
	uint8_t insertshift = 0;
	keypad = 0;

	switch(code) {
		case 0x04: maccode = 0x01; break; // A
		case 0x05: maccode = 0x17; break; // B
		case 0x06: maccode = 0x11; break; // C
		case 0x07: maccode = 0x05; break; // D
		case 0x08: maccode = 0x1D; break; // E
		case 0x09: maccode = 0x07; break; // F
		case 0x0A: maccode = 0x0B; break; // G
		case 0x0B: maccode = 0x09; break; // H
		case 0x0C: maccode = 0x45; break; // I
		case 0x0D: maccode = 0x4D; break; // J
		case 0x0E: maccode = 0x51; break; // K
		case 0x0F: maccode = 0x4B; break; // L
		case 0x10: maccode = 0x5D; break; // M
		case 0x11: maccode = 0x5B; break; // N
		case 0x12: maccode = 0x3F; break; // O
		case 0x13: maccode = 0x47; break; // P
		case 0x14: maccode = 0x19; break; // Q
		case 0x15: maccode = 0x1F; break; // R
		case 0x16: maccode = 0x03; break; // S
		case 0x17: maccode = 0x23; break; // T
		case 0x18: maccode = 0x41; break; // U
		case 0x19: maccode = 0x13; break; // V
		case 0x1A: maccode = 0x1B; break; // W
		case 0x1B: maccode = 0x0F; break; // X
		case 0x1C: maccode = 0x21; break; // Y
		case 0x1D: maccode = 0x0D; break; // Z
		case 0x1E: maccode = 0x25; break; // 1
		case 0x1F: maccode = 0x27; break; // 2
		case 0x20: maccode = 0x29; break; // 3
		case 0x21: maccode = 0x2B; break; // 4
		case 0x22: maccode = 0x2F; break; // 5
		case 0x23: maccode = 0x2D; break; // 6
		case 0x24: maccode = 0x35; break; // 7
		case 0x25: maccode = 0x39; break; // 8
		case 0x26: maccode = 0x33; break; // 9
		case 0x27: maccode = 0x3B ;break; // 0
		case 0x28: maccode = 0x49; break; // enter
		case 0x29: maccode = 0x6B; break; // ESC
		case 0x2A: maccode = 0x67; break; // bkspc
		case 0x2B: maccode = 0x61; break; // tab
		case 0x2C: maccode = 0x63; break; // space
		case 0x2D: maccode = 0x37; break; // -
		case 0x2E: maccode = 0x31; break; // =
		case 0x31: maccode = 0x55; break; // backslash

		case 0x34: maccode = 0x4F; break; // '
		case 0x35: maccode = 0x65; break; // `
		case 0x36: maccode = 0x57; break; // ,
		case 0x37: maccode = 0x5F; break; // .
		case 0x38: maccode = 0x59; break; // /
		case 0x39: maccode = 0x73; break; // caps

		case 0x55: maccode = 0x79; keypad = 0x05; insertshift = 1; break; // keypad *
		case 0x56: maccode = 0x79; keypad = 0x1D; break; // keypad -
		case 0x57: maccode = 0x79; keypad = 0x0D; insertshift = 1; break; // keypad +
		case 0x62: maccode = 0x79; keypad = 0x25; break; // keypad 0
		case 0x59: maccode = 0x79; keypad = 0x27; break; // keypad 1
		case 0x5A: maccode = 0x79; keypad = 0x29; break; // keypad 2
		case 0x5B: maccode = 0x79; keypad = 0x2B; break; // keypad 3
		case 0x5C: maccode = 0x79; keypad = 0x2D; break; // keypad 4
		case 0x5D: maccode = 0x79; keypad = 0x2F; break; // keypad 5
		case 0x5E: maccode = 0x79; keypad = 0x31; break; // keypad 6
		case 0x5F: maccode = 0x79; keypad = 0x33; break; // keypad 7
		case 0x60: maccode = 0x79; keypad = 0x37; break; // keypad 8
		case 0x61: maccode = 0x79; keypad = 0x39; break; // keypad 9
		case 0x63: maccode = 0x79; keypad = 0x03; break; // keypad .
		case 0x4F: maccode = 0x79; keypad = 0x05; break;  // right arrow
		case 0x50: maccode = 0x79; keypad = 0x0D; break;  // left arrow
		case 0x51: maccode = 0x79; keypad = 0x11; break;  // down arrow
		case 0x52: maccode = 0x79; keypad = 0x1B; break;  // up -> keypad up
		case 0x58: maccode = 0x79; keypad = 0x19; break; // keypad enter
		case 0x54: maccode = 0x79; keypad = 0x1B; break;  // keypad /
#if 0
		case 0x49: maccode = 0x72; break; // Insert/Help
		case 0x4A: maccode = 0x73; break; // Home
		case 0x4B: maccode = 0x74; break; // PG UP
		case 0x4C: maccode = 0x75; break; // Del
		case 0x4D: maccode = 0x77; break; // End
		case 0x4E: maccode = 0x79; break; // PG DN
		
#endif
	}

	if(keyup && maccode) maccode |= 0x80;

	if(maccode == 0) maccode = 0x7B;

	return (insertshift << 16) | (keypad << 8) | maccode;;
}

void MacHandleMeta(uint8_t metakey, uint8_t keyup) {
	uint32_t maccode = 0;

	switch(metakey) {
	case KBD_LEFT_SHIFT: maccode = 0x71; break;
	case KBD_RIGHT_SHIFT: maccode = 0x71; break;
	case KBD_LEFT_CTRL: maccode = 0x77; break;
	case KBD_RIGHT_CTRL: maccode = 0x77; break;
	case KBD_LEFT_GUI: maccode = 0x6F; break;
	case KBD_RIGHT_GUI: maccode = 0x6F; break;
	case KBD_LEFT_ALT: maccode = 0x75; break;
	case KBD_RIGHT_ALT: maccode = 0x75; break;
	default: return;
	}
	if(keyup) maccode |= 0x80;

	MacSendKey(maccode);
}

void MacSendKey(uint32_t maccode32) {
	uint8_t maccode;
	uint8_t insertshift;
	
	maccode = maccode32 & 0xFF;
	keypad = (maccode32 >> 8) & 0xFF;
	insertshift = (maccode32 >> 16) & 0xFF;

	__disable_irq();
	if(((key2index+1)%kBuffSize) != keyindex) {
		if((modifierkeys & kModShift) && insertshift) {
			mackey[key2index] = 0x71;
			key2index = (key2index+1) % kBuffSize;
		}
		mackey[key2index] = maccode;
		key2index = (key2index+1) % kBuffSize;
		if(keypad) {
			mackey[key2index] = keypad;
			key2index = (key2index+1) % kBuffSize;
		}
		if((modifierkeys & kModShift) && insertshift) {
			mackey[key2index] = 0xF1;
			key2index = (key2index+1) % kBuffSize;
		}
	}
	__enable_irq();
}
